/*
libutil -- stuff dealing with articles

Written by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>.
Copyright 1998.

See file COPYING for restrictions on the use of this software.
*/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <unistd.h>

#include "leafnode.h"

char * headers [10] ;	/* headers. For allocation, see leafnode.h */

/* build and return an open fd to pseudoart in group */
FILE * buildpseudoart(const char * grp) {
    FILE * f;

    f = tmpfile();
    if (!f) {
	syslog( LOG_ERR, "Could not create pseudoarticle" );
	return f;
    }

    fprintf(f, "Path: %s\n", fqdn);
    fprintf(f, "Newsgroups: %s\n", grp);
    fprintf(f, "From: Leafnode <nobody@%s>\n", fqdn);
    fprintf(f, "Subject: Leafnode placeholder for group %s\n", grp);
    fprintf(f, "Date: %s\n", rfctime() );
    fprintf(f, "Message-ID: <leafnode:placeholder:%s@%s>\n", grp, fqdn);
    fprintf(f, "\n");
    fprintf(f,
	  "This server is running leafnode, which is a dynamic NNTP proxy.\n"
	  "This means that it does not retrieve newsgroups unless someone is\n"
	  "actively reading them.\n"
	  "\n"
	  "If you do an operation on a group - such as reading an article,\n"
	  "looking at the group table of contents or similar, then leafnode\n"
	  "will go and fetch articles from that group when it next updates.\n"
	  "\n"
	  "Since you have read this dummy article, leafnode will retrieve\n"
	  "the newsgroup %s when fetch is run\n"
	  "the next time. If you'll look into this group a little later, you\n"
	  "will see real articles.\n"
	  "\n"
	  "If you see articles in groups you do not read, that is almost\n"
	  "always because of cross-posting.  These articles do not occupy any\n"
	  "more space - they are hard-linked into each newsgroup directory.\n"
	  "\n"
	  "If you do not understand this, please talk to your newsmaster.\n"
	  "\n"
	  "Leafnode can be found at\n"
	  "\thttp://wpxx02.toxi.uni-wuerzburg.de/~krasel/leafnode.html\n\n", grp );

    rewind(f);
    return f;
}

/*
 * read headers of an article into array
 * Return TRUE for success, FALSE for failure
 */
int getheaders( char * filename ) {
    FILE * f;
    int  hdr;
    char * p, *q ;

    f = fopen( filename, "r" );
    if ( !f )
	return FALSE ;

    for ( hdr = H_PATH; hdr <= H_OTHER; hdr++ )
	headers[hdr] = NULL;
    hdr = -1;

    debug = 0;
    while ( ( p = getaline( f )) && p && *p ) {
	/* allow for multiline headers */
	if ( isspace( *p ) && ( hdr >= H_PATH ) && ( hdr <= H_OTHER ) ) {
	    char *r;

	    q = p;
	    do {
		q++;
	    } while ( q && *q && isspace( *q ) );
	    r = critrealloc( headers[hdr],
			     strlen( headers[hdr] ) + strlen( q ) + 1,
			     "Allocating space for multiline headers" );
	    strcat( r, " " );
	    strcat( r, q );
	    headers[hdr] = r;
	} else {
	    q = strchr( p, ':' );
	    q++;
	    while ( isspace(*q) )
		q++;
	    if ( strncasecmp( p, "Path:", 5 ) == 0 ) {
		headers[ H_PATH ] = strdup( q );
		hdr = H_PATH ;
	    } else if ( strncasecmp( p, "Newsgroups:", 11 ) == 0 ) {
		headers[ H_NEWSGROUPS ] = strdup( q );
		hdr = H_NEWSGROUPS;
	    } else if ( strncasecmp( p, "Subject:", 8 ) == 0 ) {
		headers[ H_SUBJECT ] = strdup( q );
		hdr = H_SUBJECT ;
	    } else if ( strncasecmp( p, "From:", 5 ) == 0 ) {
		headers[ H_FROM ] = strdup( q );
		hdr = H_FROM ;
	    } else if ( strncasecmp( p, "Date:", 5 ) == 0 ) {
		headers[ H_DATE ] = strdup( q );
		hdr = H_DATE ;
	    } else if ( strncasecmp( p, "Message-ID:", 11 ) == 0 ) {
		headers[ H_MESSAGEID ] = strdup( q );
		hdr = H_MESSAGEID ;
	    } else if ( strncasecmp( p, "References:", 11 ) == 0 ) {
		headers[ H_REFERENCES ] = strdup( q );
		hdr = H_REFERENCES ;
	    } else if ( strncasecmp( p, "Lines:", 6 ) == 0 ) {
		headers[ H_LINES ] = strdup( q );
		hdr = H_LINES ;
	    } else if ( strncasecmp( p, "Xref:", 5 ) == 0 ) {
		headers[ H_XREF ] = strdup( q );
		hdr = H_XREF ;
	    } else {
		if ( !headers[ H_OTHER ] )
		    headers[ H_OTHER ] = strdup( p );	/* include header! */
		else {
		    char *r;
		    r = critrealloc( headers[ H_OTHER],
				     strlen( headers[ H_OTHER ] ) + strlen( p ) + 1,
				     "Allocating space for other headers" );
		    strcat( r, "\n" );
		    strcat( r, p );
		    headers[ H_OTHER ] = r;
		}
		hdr = H_OTHER ;
	    }
	}
    }
    debug = debugmode;

    fclose( f );
    return TRUE ;
}

/*
 * store articles in newsgroups which are already stored in
 * $SPOOLDIR/message.id/
 */
void store( const char * filename,
	    FILE * filehandle,
	    size_t bytes,
	    char * newsgroups,
	    const char * subject,
	    const char * from,
	    const char * date,
	    const char * msgid,
	    const char * references,
	    const char * lines)
{
    char tmp[10];
    static struct newsgroup * cg;
    char xrefincase[4096]; /* 1024 for newsgroups, plus article numbers */
    char * p;
    char * q;
    char * x;
    int n;

    x = xrefincase;
    n = 0;

    if ( verbose > 2 )
	printf( "storing %s: %s\n", msgid, newsgroups );

    p = newsgroups;
    while (p && *p) {
	n++;
	q = strchr( p, ',' );
	if (q)
	    *q++ = '\0';
	if ( *p ) {
	    if (!cg || strcmp(cg->name, p)) {
		cg = findgroup( p );
		if ( cg ) {
		    if ( isinteresting(cg->name) )
		        (void) chdirgroup( p, TRUE );
		    else
		    	cg = NULL;
		}
	    }
	    if ( cg ) {
		do {
		    sprintf(tmp, "%d", ++cg->last);
		    errno = 0;
		    if ( verbose > 2 )
			printf( "..as article %d in %s\n",
			       cg->last, cg->name );
		} while ( link( filename, tmp )<0 && errno==EEXIST );
		if ( errno )
		    syslog( LOG_ERR, "error linking %s into %s: %m",
			    filename, p );
		else {
		    sprintf( x, " %s:%d", cg->name, cg->last );
		    x += strlen( x );
		}
	    } else {
		if ( verbose > 2 )
		    printf( ".. discarding unknown group %s\n", p );
	    }
	}
	p = q;
    }
    if ( n > 1 )
	fprintf( filehandle, "Xref: %s%s\n", fqdn, xrefincase );
}
